#!/usr/bin/env python

"""
Pandoc filter to process raw latex tikz environments into images.
Assumes that pdflatex is in the path, and that the standalone
package is available.  Also assumes that ImageMagick's convert
is in the path. Images are put in the tikz-images directory.

CREDIT: Did I write this? Or downloaded it from somewhere? Oh the mystery.
"""

import os
import re
import shutil
import sys
from subprocess import call
from tempfile import mkdtemp

from pandocfilters import toJSONFilter, Para, Image, get_filename4code, get_extension


def tikz2image(tikz_src, filetype, outfile):
    tmpdir = mkdtemp()
    olddir = os.getcwd()
    os.chdir(tmpdir)
    f = open('tikz.tex', 'w')
    f.write("""\\documentclass{standalone}
             \\usepackage{tikz}
             \\begin{document}
             """)
    f.write(tikz_src)
    f.write("\n\\end{document}\n")
    f.close()
    call(["pdflatex", 'tikz.tex'], stdout=sys.stderr)
    os.chdir(olddir)
    if filetype == 'pdf':
        shutil.copyfile(tmpdir + '/tikz.pdf', outfile + '.pdf')
    else:
        call(["convert", tmpdir + '/tikz.pdf', outfile + '.' + filetype])
    shutil.rmtree(tmpdir)


def tikz(key, value, format, _):
    if key == 'RawBlock':
        [fmt, code] = value
        if fmt == "latex" and re.match("\\\\begin{tikzpicture}", code):
            outfile = get_filename4code("tikz", code)
            filetype = get_extension(format, "png", html="png", latex="pdf")
            src = outfile + '.' + filetype
            if not os.path.isfile(src):
                tikz2image(code, filetype, outfile)
                sys.stderr.write('Created image ' + src + '\n')
            return Para([Image(['', [], []], [], [src, ""])])

if __name__ == "__main__":
    toJSONFilter(tikz)



"""logger.py: 
Logging support.
"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import os
import logging

def init_logger( filename = '_log.txt', name='root', level = 'info'):
    logging.basicConfig( level=logging.DEBUG,
            format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            datefmt='%m-%d %H:%M',
            filename= filename,
            filemode='a'
            )
    console = logging.StreamHandler()
    console.setLevel( eval('logging.%s'%level.upper() ))
    formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
    console.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.addHandler(console)
    return logger

logger = init_logger()

def main():
    logger.info( 'hey wassup?' )
    logger.debug( 'Nothing much. Few bits!' )

if __name__ == '__main__':
    main()



"""text_processing.py: 

Text processing functions.

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2016, Dilawar Singh"
__credits__          = ["NCBS Bangalore"]
__license__          = "GNU GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import re

float_regex_ = re.compile( r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?' )

def to_float( token ):
    val = token
    try:
        val = float( token )
    except Exception as e:
        pass
    return val

def remove_special_char( txt ):
    return re.sub( r'_|\+|\-|\%', ' ', txt )

def find_floats( text ):
    """Find all floats in given string
    """
    global float_regex_
    assert type( text ) == str
    return [ to_float( m.group(0) ) for m  in float_regex_.finditer( text ) ]

def arrays2csv( outfile, **kwargs ):
    """Write given vector/array to a CSV file. The separator is ',' by default.
    It can not be changed. It can not be changed. It can not be changed.

    The argument order will be preserved if python-3.6+ is used. Not
    implementing the order preservation here.
    """
    header = map( remove_special_char, kwargs.keys( ) )
    sep = ','
    with open( outfile, 'w' ) as f:
        f.write( sep.join( header ) + '\n' )
        cols = kwargs.values( )
        for x in zip( *cols ):
            x = map( str, x )
            f.write( sep.join( x ) + '\n' )


# deprecated
def find_all_floats( txt ):
    return find_floats( txt )



# -*- coding: utf-8 -*-
"""
Utility for brian2 simulator.
"""

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

from brian2 import *
import pandas as pd

def spikeMonitor2DataFrame(sm):
    df = pd.DataFrame()
    trains = sm.spike_trains()
    df = pd.DataFrame({k: pd.Series(v) for k, v in trains.items()})
    return df

def stateMonitorToDataFrame(sm):
    df = pd.DataFrame()
    df['t'] = np.array(sm.t/second)
    for x in sm.record_variables:
        vals = getattr(sm, x)
        for i, v in enumerate(vals):
            vname = f'{x}[{i}]'
            df[vname] = v
    return df

def test():
    N = 100
    taum = 10*ms
    taupre = 20*ms
    taupost = taupre
    Ee = 0*mV
    vt = -54*mV
    vr = -60*mV
    El = -74*mV
    taue = 5*ms
    F = 15*Hz
    gmax = .01
    dApre = .01
    dApost = -dApre * taupre / taupost * 1.05
    dApost *= gmax
    dApre *= gmax

    eqs_neurons = '''
    dv/dt = (ge * (Ee-vr) + El - v) / taum : volt
    dge/dt = -ge / taue : 1
    '''

    input = PoissonGroup(N, rates=F)
    neurons = NeuronGroup(1, eqs_neurons, threshold='v>vt', reset='v = vr',
                          method='linear')
    S = Synapses(input, neurons,
                 '''w : 1
                    dApre/dt = -Apre / taupre : 1 (event-driven)
                    dApost/dt = -Apost / taupost : 1 (event-driven)''',
                 on_pre='''ge += w
                        Apre += dApre
                        w = clip(w + Apost, 0, gmax)''',
                 on_post='''Apost += dApost
                         w = clip(w + Apre, 0, gmax)''',
                 )
    S.connect()
    S.w = 'rand() * gmax'
    mon = StateMonitor(S, 'w', record=[0, 1])
    s_mon = SpikeMonitor(input)
    run(10*second, report='text')
    df1 = spikeMonitor2DataFrame(s_mon)
    df2 = stateMonitorToDataFrame(mon)
    ax1 = plt.subplot(211)
    df1.plot(ax=ax1)
    ax2 = plt.subplot(212)
    df2.plot(x='t', ax=ax2)
    tight_layout()
    savefig(f'{__file__}.png')

def main():
    test()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""
Pandoc filter to convert divs with class="theorem" to LaTeX
theorem environments in LaTeX output, and to numbered theorems
in HTML output.
"""

from pandocfilters import toJSONFilter, RawBlock, Div

theoremcount = 0


def latex(x):
    return RawBlock('latex', x)


def html(x):
    return RawBlock('html', x)

def theorems(key, value, format, meta):
    if key == 'Div':
        [[ident, classes, kvs], contents] = value
        if "theorem" in classes:
            if format == "latex":
                if ident == "":
                    label = ""
                else:
                    label = '\\label{' + ident + '}'
                return([latex('\\begin{theorem}' + label)] + contents +
                       [latex('\\end{theorem}')])
            elif format == "html" or format == "html5":
                global theoremcount
                theoremcount = theoremcount + 1
                newcontents = [html('<dt>Theorem ' + str(theoremcount) + '</dt>'),
                               html('<dd>')] + contents + [html('</dd>\n</dl>')]
                return Div([ident, classes, kvs], newcontents)

if __name__ == "__main__":
    toJSONFilter(theorems)



#!/usr/bin/env python

"""
Pandoc filter to allow interpolation of metadata fields
into a document.  %{fields} will be replaced by the field's
value, assuming it is of the type MetaInlines or MetaString.
"""

from pandocfilters import toJSONFilter, attributes, Span, Str
import re

pattern = re.compile('%\{(.*)\}$')


def metavars(key, value, format, meta):
    if key == 'Str':
        m = pattern.match(value)
        if m:
            field = m.group(1)
            result = meta.get(field, {})
            if 'MetaInlines' in result['t']:
                return Span(attributes({'class': 'interpolated',
                                        'field': field}),
                            result['c'])
            elif 'MetaString' in result[t]:
                return Str(result['c'])

if __name__ == "__main__":
    toJSONFilter(metavars)



#!/usr/bin/env python
from pandocfilters import toJSONFilter, RawInline

"""
Pandoc filter that causes emphasis to be rendered using
the custom macro '\myemph{...}' rather than '\emph{...}'
in latex.  Other output formats are unaffected.
"""


def latex(s):
    return RawInline('latex', s)


def myemph(k, v, f, meta):
    if k == 'Emph' and f == 'latex':
        return [latex('\\myemph{')] + v + [latex('}')]

if __name__ == "__main__":
    toJSONFilter(myemph)



# -*- coding: utf-8 -*-
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

# These are helper functions to store metadata to HDF5.
# Snippets originally from https://stackoverflow.com/a/29130146/1805129
def h5store(filename, df, key = 'mydata', **kwargs):
    import pandas as pd
    store = pd.HDFStore(filename, mode=kwargs.get('mode', 'w'))
    store.put(key, df)
    store.get_storer(key).attrs.metadata = kwargs
    store.close()

def h5load(filename, key='mydata'):
    import pandas as pd
    store = pd.HDFStore(filename)
    data = store[key]
    metadata = store.get_storer(key).attrs.metadata
    return data, metadata

save_df = h5store
load_df = h5load

def tests():
    import os
    import numpy as np
    import pandas as pd
    import tempfile
    storeName = os.path.join(tempfile.gettempdir(),'data.h5')
    data  = np.random.rand(100, 100)
    df = pd.DataFrame(data)
    h5store(storeName, df, me='dilawar', mynum = 301810)
    s, m = h5load(storeName)
    assert df.equals(s)
    assert m == dict(me ='dilawar', mynum=301810)

def main():
    tests()

if __name__ == '__main__':
    main()



"""file_utils.py: 

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import os
import subprocess


def tail(f, n, offset=0):
    """Read n lines from the back of file.
    Work on UNIX systems only.
    """
    lines = subprocess.check_output( "tail -n %d %s" % (n+offset, f ), shell =
            True )
    return lines.decode( 'utf-8' )



#!/usr/bin/env python3
"""
Pandoc filter to convert code blocks with class "gabc" to LaTeX
\\gabcsnippet commands in LaTeX output, and to images in HTML output.
Assumes Ghostscript, LuaLaTeX, [Gregorio](http://gregorio-project.github.io/)
and a reasonable selection of LaTeX packages are installed.
"""

import os
from sys import getfilesystemencoding, stderr
from subprocess import Popen, call, PIPE, DEVNULL
from hashlib import sha1
from pandocfilters import toJSONFilter, RawBlock, RawInline, Para, Image


IMAGEDIR = "tmp_gabc"
LATEX_DOC = """\\documentclass{article}
\\usepackage{libertine}
\\usepackage[autocompile]{gregoriotex}
\\pagestyle{empty}
\\begin{document}
%s
\\end{document}
"""


def sha(code):
    """Returns sha1 hash of the code"""
    return sha1(code.encode(getfilesystemencoding())).hexdigest()


def latex(code):
    """LaTeX inline"""
    return RawInline('latex', code)


def latexblock(code):
    """LaTeX block"""
    return RawBlock('latex', code)


def htmlblock(code):
    """Html block"""
    return RawBlock('html', code)


def latexsnippet(code, kvs):
    """Take in account key/values"""
    snippet = ''
    staffsize = int(kvs['staffsize']) if 'staffsize' in kvs else 17
    annotationsize = .56 * staffsize
    if 'mode' in kvs:
        snippet = (
            "\\greannotation{{\\fontsize{%s}{%s}\\selectfont{}%s}}\n" %
            (annotationsize, annotationsize, kvs['mode'])
        ) + snippet
    if 'annotation' in kvs:
        snippet = (
            "\\grechangedim{annotationseparation}{%s mm}{0}\n"
            "\\greannotation{{\\fontsize{%s}{%s}\\selectfont{}%s}}\n" %
            (staffsize / 34, annotationsize, annotationsize, kvs['annotation'])
        ) + snippet
    snippet = (
        "\\grechangestaffsize{%s}\n" % staffsize +
        "\\def\\greinitialformat#1{{\\fontsize{%s}{%s}\\selectfont{}#1}}" %
        (2.75 * staffsize, 2.75 * staffsize)
    ) + snippet
    snippet = "\\setlength{\\parskip}{0pt}\n" + snippet + code
    return snippet


def latex2png(snippet, outfile):
    """Compiles a LaTeX snippet to png"""
    pngimage = os.path.join(IMAGEDIR, outfile + '.png')
    environment = os.environ
    environment['openout_any'] = 'a'
    environment['shell_escape_commands'] = \
        "bibtex,bibtex8,kpsewhich,makeindex,mpost,repstopdf,gregorio"
    proc = Popen(
        ["lualatex", '-output-directory=' + IMAGEDIR],
        stdin=PIPE,
        stdout=DEVNULL,
        env=environment
    )
    proc.stdin.write(
        (
            LATEX_DOC % (snippet)
        ).encode("utf-8")
    )
    proc.communicate()
    proc.stdin.close()
    call(["pdfcrop", os.path.join(IMAGEDIR, "texput.pdf")], stdout=DEVNULL)
    call(
        [
            "gs",
            "-sDEVICE=pngalpha",
            "-r144",
            "-sOutputFile=" + pngimage,
            os.path.join(IMAGEDIR, "texput-crop.pdf"),
        ],
        stdout=DEVNULL,
    )


def png(contents, latex_command):
    """Creates a png if needed."""
    outfile = sha(contents + latex_command)
    src = os.path.join(IMAGEDIR, outfile + '.png')
    if not os.path.isfile(src):
        try:
            os.mkdir(IMAGEDIR)
            stderr.write('Created directory ' + IMAGEDIR + '\n')
        except OSError:
            pass
        latex2png(latex_command + "{" + contents + "}", outfile)
        stderr.write('Created image ' + src + '\n')
    return src


def gabc(key, value, fmt, meta):                   # pylint:disable=I0011,W0613
    """Handle gabc file inclusion and gabc code block."""
    if key == 'Code':
        [[ident, classes, kvs], contents] = value  # pylint:disable=I0011,W0612
        kvs = {key: value for key, value in kvs}
        if "gabc" in classes:
            if fmt == "latex":
                if ident == "":
                    label = ""
                else:
                    label = '\\label{' + ident + '}'
                return latex(
                    "\n\\smallskip\n{%\n" +
                    latexsnippet('\\gregorioscore{' + contents + '}', kvs) +
                    "%\n}" +
                    label
                )
            else:
                infile = contents + (
                    '.gabc' if '.gabc' not in contents else ''
                )
                with open(infile, 'r') as doc:
                    code = doc.read().split('%%\n')[1]
                return [Image(['', [], []], [], [
                    png(
                        contents,
                        latexsnippet('\\gregorioscore', kvs)
                    ),
                    ""
                ])]
    elif key == 'CodeBlock':
        [[ident, classes, kvs], contents] = value
        kvs = {key: value for key, value in kvs}
        if "gabc" in classes:
            if fmt == "latex":
                if ident == "":
                    label = ""
                else:
                    label = '\\label{' + ident + '}'
                return [latexblock(
                    "\n\\smallskip\n{%\n" +
                    latexsnippet('\\gabcsnippet{' + contents + '}', kvs) +
                    "%\n}" +
                    label
                    )]
            else:
                return Para([Image(['', [], []], [], [
                    png(
                        contents,
                        latexsnippet('\\gabcsnippet', kvs)
                    ),
                    ""
                ])])


if __name__ == "__main__":
    toJSONFilter(gabc)



#!/usr/bin/env python

"""
Pandoc filter to process code blocks with class "plantuml" into
plant-generated images.

Needs `plantuml.jar` from http://plantuml.com/.
"""

import os
import sys
from subprocess import call

from pandocfilters import toJSONFilter, Para, Image, get_filename4code, get_caption, get_extension


def plantuml(key, value, format, _):
    if key == 'CodeBlock':
        [[ident, classes, keyvals], code] = value

        if "plantuml" in classes:
            caption, typef, keyvals = get_caption(keyvals)

            filename = get_filename4code("plantuml", code)
            filetype = get_extension(format, "png", html="svg", latex="eps")

            src = filename + '.uml'
            dest = filename + '.' + filetype

            if not os.path.isfile(dest):
                txt = code.encode(sys.getfilesystemencoding())
                if not txt.startswith("@start"):
                    txt = "@startuml\n" + txt + "\n@enduml\n"
                with open(src, "w") as f:
                    f.write(txt)

                call(["java", "-jar", "plantuml.jar", "-t"+filetype, src])
                sys.stderr.write('Created image ' + dest + '\n')

            return Para([Image([ident, [], keyvals], caption, [dest, typef])])

if __name__ == "__main__":
    toJSONFilter(plantuml)



"""statistics.py: 
"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

def histogram( vec, bins = 10, normed = False ):
    """Histogram of given vector.
    """
    a, b = min(vec), max(vec)
    s = (b-a) / float( bins )
    buckets = [ a + s * x for x in range( bins ) ]
    buckets.append( b )

    hist = [ 0 ] * bins
    for i, bb in enumerate( buckets[1:] ):
        aa = buckets[i]
        for x in vec:
            if x >= aa and x <= bb:
                hist[i] += 1

    if normed:
        hist = [ float(x) / len( vec ) for x in hist ]

    return hist, buckets

def test( ):
    import numpy as np
    data = np.random.random( 100 ) * 100
    data = [10,20,30,40,50,60,70,80,90,100]
    print( histogram( data ) )
    print( np.histogram( data ) )

if __name__ == '__main__':
    test()



__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

import os
import re
import sys
import time
import subprocess
import colorama as C
from pathlib import Path

C.init(autoreset=True)

sdir_ = Path(__file__).parent

all_ = [ 'pandoc-imagine'
        , sdir_ / 'gls.py'  # Don't know why but this should come before cross-ref
        , 'pandoc-crossref'
        , 'pandoc-citeproc' 
        , 'pandoc-imagine' 
        , 'pantable'
        , sdir_ / 'quantity.py'
        ]

# This is from  https://stackoverflow.com/a/377028/1805129
def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


def available_pandoc_filters():
    global all_
    cmds = [ which(prog) for prog in all_]
    return [str(x) for x in cmds if x is not None]

def _prefix(msg):
    return f"{C.Fore.BLUE}dilawar.pandoc>{C.Style.RESET_ALL}{msg.strip()}"

def executeCommand(cmd):
    print(_prefix(f"Executing {cmd}"), file=sys.stderr)
    cmd = cmd.split()
    p = subprocess.Popen(cmd
            , stdout=subprocess.PIPE
            , stderr=subprocess.STDOUT
            , stdin=sys.stdin
            , universal_newlines=True
            , encoding="utf-8"
            )
    while True:
        line = p.stdout.readline()
        if not line:
            break
        print(_prefix(line), file=sys.stderr)
    return p.returncode

def execute_pandoc(*args):
    """Top level command.
    """
    #  sdir = Path(__file__).parent
    #  css = sdir / 'pandoc.css'
    argStr = ' '.join(*args)
    extra = ''
    pandoc = which('pandoc')
    if re.search(r'-t\s+latex', argStr):
        extra += ' --pdf-engine lualatex -s'
    if re.search(r'-t\s+html\S*', argStr):
        # --katex may not work when we are behing proxy. So let USER specify it.
        extra += ' --self-contained'
        ##if re.search(r'--css\s+', argStr) is None:
        ##    extra += f' --css {css} '
    filters = ' '.join([f'-F {f}' for f in available_pandoc_filters()])
    cmd = f'{pandoc} {filters} {extra} ' + argStr
    t0 = time.time()
    st = executeCommand(cmd)
    print(_prefix(f" Took {time.time()-t0:.2f} sec"), file=sys.stderr)
    return st

def t_pandoc():
    # test pandoc.
    infile = Path(__file__).parent /'..'/'..'/'README.md'
    assert infile.exists(), infile
    execute_pandoc(f'-t html5 -o a.html {infile}'.split())
    

def test():
    t_pandoc()

if __name__ == '__main__':
    test()



"""information_theory.py: 

Functions related to information theory.

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import math


def entropy( P ):
    h = 0.0
    for p in P:
        assert p >= 0, 'All values must be positive'
        if p == 0:
            continue
        h -= p * math.log( p, 2 )
    return h


def kl_distance( P, Q ):
    dist = 0.0
    for p, q in zip( P, Q ):
        assert q > 0, 'Q must not have zero entry'
        if p == 0:
            continue 
        dist += (p  * math.log( p/q, 2 ))
    return dist



#!/usr/bin/env python3
"""
Pandoc filter to process code blocks with class "graphviz", "chemfig",
"standalone" into pdf images.

Dependencies:
    - pygraphviz
    - lualatex
"""

import os
import sys
import re
import subprocess

from pandocfilters import toJSONFilter, Para, Str, Strong, RawInline
from pandocfilters import Image, get_filename4code, get_caption, get_extension
from pandocfilters import RawBlock, CodeBlock


def latex( x ):
    return RawBlock( 'latex',  x)

def print1( *msg ):
    print( ' | '.join(msg), file=sys.stderr )

def replace_ext( filename, newext = 'tex' ):
    oldExt = filename.split( '.' )[-1]
    return re.sub( r'(.+?)\.%s$' % oldExt, '\1\.%s' % newext )

def replace_relative_filepaths( code ):
    # Usually a datafile for gnuplot or table. It may not be surrounded by \".
    # Well sniff all paths which starts with ./ and prefix them with .. since
    # these are one level up.
    filePat = re.compile( r'\.\/((.+?)\.(csv|dat|txt))', re.I )

    # We can't replace using simple str.replace function since same filepath is
    # often repeated in code. We need to find the location.
    newText, start = [], 0
    for m in filePat.finditer( code ):
        filepath = m.group(0)
        a, b = m.span()
        replaceWith = '../%s' % filepath
        newText.append( code[start:a] )
        newText.append( replaceWith )
        start = b

    newText.append( code[start:] )
    code = ''.join( newText )
    #  code = code.replace( filepath, '../%s' % filepath )
    return code

def gen_standalone( code, dest ):
    dest = os.path.realpath( dest )
    ext = dest.split( '.' )[-1]

    code = replace_relative_filepaths( code ) 

    code = r'%s' % code

    tex = [ '\\RequirePackage{luatex85,shellesc}' ]
    tex += [ '\\documentclass[preview,multi=false]{standalone}' ]
    tex += [ '\\usepackage{amsmath,amssymb}' ]
    tex += [ '\\usepackage[sfdefault]{FiraSans}' ]
    tex += [ '\\usepackage[small,euler-digits]{eulervm}' ]
    tex += [ '\\usepackage{chemfig}' ]
    tex += [ '\\usepackage{tikz}' ]
    tex += [ '\\usetikzlibrary{calc,shapes,arrows, arrows.meta, positioning,fit}' ]
    tex += [ '\\usepackage{pgfplots}' ]
    tex += [ '\\usepgfplotslibrary{units}' ]
    if r'\begin{document}' not in code:
        tex += [ '\\begin{document}' ]
    tex += [ code ]
    if r'\end{document}' not in code:
        tex += [ '\\end{document}']

    dirname = os.path.dirname( dest )
    basename = os.path.basename( dest )
    nameWE = '.'.join( basename.split( '.' )[:-1] )
    texFile = os.path.join( dirname, nameWE + '.tex' )

    # Write file
    with open( texFile, 'w' ) as f:
        f.write( '\n'.join( tex ) )

    res1 = subprocess.run( [ 'lualatex', '-shell-escape', texFile ]
            #  , shell=False, stderr = subprocess.STDOUT
            , cwd = dirname
            , check = False
            , stdout = subprocess.PIPE, stderr = subprocess.PIPE
            )

    if res1.returncode != 0:
        print1( "WARN", "It seems previous command failed." )
        print1( '%s' % res1.stdout.decode( 'utf8' )  )
        return False

    # destination may not be pdf. Use imagick to convert the pdf to other format.
    if ext != 'pdf':
        pdfFile = os.path.join( dirname, nameWE + '.pdf' )
        outfile = os.path.join( dirname, nameWE + '.%s' % ext )
        opts = '-density 300 -antialias -quality 100'. split( )
        res = subprocess.run( 
                [ 'convert', pdfFile ] + opts + [ outfile ]
                , shell=False
                , stdout = subprocess.PIPE, stderr = subprocess.STDOUT
                , cwd = dirname
                )

    if not os.path.isfile( dest ):
        print1( "ERROR", "%s could not be generated." % dest )
        return False

    return True

def codeblocks(key, value, format, _):
    if key == 'CodeBlock':
        return process( value, format )

def process( value, format ):
    [[ident, classes, keyvals], code] = value

    if "graphviz" in classes:
        caption, typef, keyvals = get_caption(keyvals)
        filetype = get_extension(format, "png", html="png", latex="pdf")
        dest = get_filename4code("graphviz", code, filetype)

        if not os.path.isfile(dest):
            import pygraphviz
            g = pygraphviz.AGraph(string=code)
            g.layout()
            g.draw(dest)
            print1('INFO', 'Created image ' + dest + '\n')

        return Para([Image([ident, [], keyvals], caption, [dest, typef])])

    elif "standalone" in classes:
        caption, typef, keyvals = get_caption(keyvals)
        filetype = get_extension(format, "png", html="png", latex="pdf")
        dest = get_filename4code("standalone", code, filetype)
        if not os.path.isfile(dest):
            success = gen_standalone(code, dest)
            if not success:
                return Para([ Str(">>> Error: This image could not be generated.")] )
        return Para([Image([ident, [], keyvals], caption, [dest, typef])])

    elif "include" in classes:
        caption, typef, keyvals = get_caption(keyvals)
        keyvalDict = dict(keyvals)

        listing = ''
        with open( keyvalDict['src'], 'r' ) as f:
            listing = f.read() 

        lang = keyvalDict.get( 'lang', 'python')
        code = 'Failed to find any listing.'
        if 'start' in keyvalDict:
            # Assume that both start and end are line numbers.
            start = int( keyvalDict['start'] )
            end = int( keyvalDict['end'] )
            code = '\n'.join( listing.split('\n')[start:end] )
        elif 'pat' in keyvalDict:
            pat = r'%s' % keyvalDict['pat']
            print1( pat )
            m = re.search( pat, listing, re.DOTALL )
            if m:
                code = m.group(0)
            else:
                code = "Pattern '%s' not found in '%s'" % (pat, keyvalDict['src'])
        else:
            code = 'No listing found.'
        return CodeBlock([ident, [], keyvals], code)

if __name__ == "__main__":
    toJSONFilter( codeblocks )



#!/usr/bin/env python

"""
Pandoc filter to process code blocks with class "abc" containing
ABC notation into images.  Assumes that abcm2ps and ImageMagick's
convert are in the path.  Images are put in the abc-images directory.
"""

import os
import sys
from subprocess import Popen, PIPE, call

from pandocfilters import toJSONFilter, Para, Image, get_caption, get_filename4code, get_extension


def abc2eps(abc_src, filetype, outfile):
    p = Popen(["abcm2ps", "-O", outfile + '.eps', "-"], stdin=PIPE)
    p.stdin.write(abc_src)
    p.communicate()
    p.stdin.close()
    call(["convert", outfile + '.eps', outfile + '.' + filetype])


def abc(key, value, format, _):
    if key == 'CodeBlock':
        [[ident, classes, keyvals], code] = value
        if "abc" in classes:
            caption, typef, keyvals = get_caption(keyvals)
            outfile = get_filename4code("abc", code)
            filetype = get_extension(format, "png", html="png", latex="pdf")
            dest = outfile + '.' + filetype

            if not os.path.isfile(dest):
                abc2eps(code.encode("utf-8"), filetype, outfile)
                sys.stderr.write('Created image ' + dest + '\n')

            return Para([Image([ident, [], keyvals], caption, [dest, typef])])


if __name__ == "__main__":
    toJSONFilter(abc)



"""ajgar.py: Main module.

"""
from __future__ import print_function, division, absolute_import
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2016, Dilawar Singh"
__license__          = "GNU GPL"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

try:
    # This depends on scipy
    from dilawar.data_analysis import *
except ImportError:
    pass

from dilawar.text_processing import *
from dilawar.file_utils import *
from dilawar.plot_utils import *
from dilawar.statistics import *
from dilawar.io_utils import *
from dilawar.logger import *
from dilawar.information_theory import *
from dilawar.functions import *

try:
    import networkx as nx
    from dilawar.nx_utils import *
except Exception as e:
    pass

# NOTE: brian2 should not be imported by default. 



# Helpful functions.

__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

import numpy as np

# This one is from here:
# https://scipy-cookbook.readthedocs.io/items/Matplotlib_SigmoidalFunctions.html
def boltzman(x, xmid, tau):
    """
    evaluate the boltzman function with midpoint xmid and time constant tau
    over x
    """
    return 1. / (1. + np.exp(-(x-xmid)/tau))

def hill(x, n, k=1):
    """hill function.
    """
    return k*(x**n/(1+x**n))

def logistic(x, alpha=1):
    """logistic function.
    """
    return (1+np.exp(-x))**(-alpha)

def sigmoid(x):
    """Sigmoid function.

    """
    return logistic(x, 1)


def test():
    # Plot all functions.
    import matplotlib.pyplot as plt
    x = np.linspace(0, 1, 1000)
    plt.subplot(221)
    for tau in [0.05, 0.1, 0.2, 0.4, 0.8]:
        y = boltzman(x, 0.5, tau)
        plt.plot(x, y, label=f'{tau}')
    plt.title(r'Botzman function: $\frac{1}{1+e^{-(x-x_o)/τ}}$')
    plt.legend()

    plt.subplot(222)
    x = np.logspace(-2, 2, 1000)
    for n in [1,2,3,4]:
        y = hill(x, n)
        plt.semilogx(x, y, label=f'n={n}')
    plt.legend()
    plt.title(r'Hill function: $\frac{x^n}{1+x^n}$')

    plt.subplot(223)
    x = np.linspace(-5, 5, 1000)
    for alpha in [0, 0.1, 0.5, 1, 2, 4, 10, 100]:
        plt.plot(x, logistic(x, alpha), label=fr'$α={alpha}$')
    plt.legend(bbox_to_anchor=(1,1), loc='upper left', fontsize=8)
    plt.title(r'Logistic: $(1+e^{-x})^{-α}$')


    plt.tight_layout()
    plt.show()

if __name__ == '__main__':
    test()



"""
Convert a pandoc file to PDF.
"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import os
import re
from pathlib import Path
import subprocess 
import dilawar.pandoc.utils as panu

srcFile_ = None
sdir_ = Path(__file__).parent

latexBin = panu.which('lualatex') or panu.which('pdflatex')
assert latexBin is not None, "No lualatex or pdflatex found in your system's PATH" 

pandoc_ = [ 'pandoc', '--to', 'latex'
        , '-s'  # standalone, in case of TeX
        , '-N'  # Numbered section.
        , '--pdf-engine', latexBin ]

# append pandoc filters.
for cmd in panu.available_pandoc_filters():
    pandoc_.append('-F')
    pandoc_.append(cmd)

# imgPat_ = re.compile( r'\!\[.*?\]\(?P<filename>.?\)' )
imgPat_ = re.compile( r'\!\[.*?\]\((?P<figure>.+?)\)', re.DOTALL )

def convertToPNG( img, text ):
    global srcFile_
    srcDir = os.path.dirname( srcFile_ )
    imgPath = os.path.join( srcDir, img )
    imgNameWe = '.'.join( img.split( '.' )[:-1] )
    pngPath = imgNameWe + '.png'
    if os.path.isfile( pngPath ):
        # File exits, if modification of png is later than the img then ignore.
        if os.path.getctime( pngPath ) > os.path.getctime( img ):
            text = text.replace( img, pngPath )
            return text

    print( 'Converting %s to %s' % (imgPath, pngPath) )
    cmd = [ 'convert', '-density', '300', '-quality', '90', imgPath, pngPath ]
    subprocess.check_output(cmd, shell = False )
    if not os.path.exists( pngPath ):
        print( '[WARN] to create PNG file' )
        print( '\t CMD: %s' % ' '.join( cmd ) )
    else:
        text = text.replace( img, pngPath )
    return text

def toPDF(text, outfile):
    global srcFile_
    global pandoc_
    pFile = '%s_pdf.md' % srcFile_ 
    with open(pFile, 'w') as f:
        f.write(text)
    cmd = pandoc_ + [ '-o', outfile, pFile ]
    subprocess.call( cmd, shell = False )
    print( '[INFO] Wrote PDF/TeX to %s' % outfile )


def process(text, outfile):
    imgs = imgPat_.findall( text )
    for img in imgs:
        ext = img.split( '.' )[-1]
        if ext.lower( ) not in [ 'png', 'jpg', 'jpeg', 'pdf', 'eps', 'gif']:
            text = convertToPNG( img, text )

    with open( '/tmp/%s_tex' % os.path.basename( srcFile_ ), 'w' ) as f:
        f.write( text )
    toPDF(text, outfile)

def main(args):
    global srcFile_
    srcFile_ = args.input
    with open( srcFile_, 'r' ) as f:
        text = f.read( )
    outfile = args.output or f'{srcFile_}.pdf'
    process(text, outfile)

if __name__ == '__main__':
    import argparse
    # Argument parser.
    description = '''Convert md to PDF'''
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('input', help = 'Input md file')
    parser.add_argument('--output', '-o'
            , required = False
            , help = 'Output file'
            )
    class Args: pass 
    args = Args()
    parser.parse_args(namespace=args)
    main(args)



#!/usr/bin/env python3

"""
A filter for parsing quantities expressed inside \Q or \quantity macro e.g.
\Q{1 m/s}, \quantity{9.8 m per second^2} etc.

Requires:
    - pip3 install pint
    - pip3 install panflute
"""

import re
import panflute as P
from pint import UnitRegistry
import itertools

U_ = UnitRegistry()

def searchQuantity(text):
    text = text.strip()
    return [x for x in 
            re.finditer(r'(\\Q|\\quantity)\{(?P<val>\S+)\s+(?P<uexpr>[^}]+)\}'
                , text)]

def reformatElem(elem, fmt='md'):
    num, unit = elem.text.split(' ', 2)
    numF = formatEval(num, fmt)
    txt = f'{numF} {unit}'
    new = P.convert_text(txt)[0].content
    return list(new)


def formatEval(val, fmt):
    # TODO: Add a cheap test to check if val can be converted to float.
    val = val.lower()
    val = re.sub(r'(1(\.[0]*)?)e', r'e', val)
    if 'e' not in val:
        return val
    a, b = val.split('e')

    # This block return 1e-20 as e-20 to latex and 10~-20~ to markdown. Prefix
    # is removed to save space.
    if not a:
        if fmt == 'latex':
            return f'10^{b}'
        else:
            return f'10^{b}^'

    if fmt == 'latex':
        return val
    else:
        return f'{a}^{b}^'

def formatQuantity(qexpr, fmt):
    ms = searchQuantity(qexpr)
    res = []
    if not ms:
        res.append((None, qexpr, False))
        return res
    for m in ms:
        val, uexpr = m.groupdict()['val'], m.groupdict()['uexpr']
        try:
            val = U_(f'{val} {uexpr}')
            res.append((m, val,True))
        except Exception as e:
            P.debug(f'[WARN] Failed to parse {qexpr}. Error: {e}')
            res.append((m, qexpr, False))
    return res

def action_quantity(elem, doc):
    w_ = doc.format
    if isinstance(elem, P.RawInline):
        # Here we have simple replacement of whole string.
        for m, f, success in formatQuantity(elem.text, w_):
            if success:
                if w_ == 'latex':
                    elem.text = f'{f:Lx}'
                else:
                    elem.text = f'{f:~P}'
                    return reformatElem(elem)
    elif isinstance(elem, P.Math) or isinstance(elem, P.RawBlock):
        # Here we have to replace part of the string. A bit more complicated
        # than before. Use gonna use m.span() to find the locations.
        toreplace = []
        for m, f, success in formatQuantity(elem.text, w_):
            if success:
                toreplace.append((m.span(), f))

        noMatch, new = [], []
        prevI, b = 0, 0
        for (a, b), f in toreplace:
            noMatch.append(elem.text[prevI:a])
            f = f'{f:Lx}' if w_ == 'latex' else f'{f:~P}'
            new.append(f)
            prevI = b
        # rest of the text
        noMatch.append(elem.text[b:])
        if toreplace:
            newtext = ''.join(itertools.chain(noMatch, new))
            if elem.text != newtext:
                elem.text = newtext

def main(doc=None):
    P.run_filter(action_quantity, doc=doc)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Pandoc filter to convert definition lists to bullet
lists with the defined terms in strong emphasis (for
compatibility with standard markdown).
"""

from pandocfilters import toJSONFilter, BulletList, Para, Strong


def deflists(key, value, format, meta):
    if key == 'DefinitionList':
        return BulletList([tobullet(t, d) for [t, d] in value])


def tobullet(term, defs):
    return([Para([Strong(term)])] + [b for d in defs for b in d])


if __name__ == "__main__":
    toJSONFilter(deflists)



#!/usr/bin/env python

"""
Pandoc filter to convert all regular text to uppercase.
Code, link URLs, etc. are not affected.
"""

from pandocfilters import toJSONFilter, Str


def caps(key, value, format, meta):
    if key == 'Str':
        return Str(value.upper())

if __name__ == "__main__":
    toJSONFilter(caps)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Replace \gls to their value in TeX file. It read the \loadglsentries files from
Tex File to find the replacement. 

Replace \SI and \si etc as well.

First encounter is replaced by large form, rest are replaced by short form.

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2017-, Dilawar Singh"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"

import sys
import os
import re

def find_glossaries( tex, dirname ):
    pat = re.compile( r'\\loadglsentries\{\s*(?P<fname>.+?)\}' )
    return [ os.path.join( dirname, x) for x in pat.findall( tex )] 

def read_glossary( f, gls ):
    if not os.path.isfile( f ):
        f = '%s.tex' % f

    # If glossary file is not found, don't do anything.
    if not os.path.exists(f):
        return 

    glpat = re.compile( r'\\newacronym\{\s*(?P<gls>.+?)\}' + \
            r'\s*\{(?P<short>.+?)\}\s*\{(?P<long>.+?)\}', re.DOTALL )

    with open( f, 'r' ) as h:
        glsText = h.read( )
        for m in glpat.findall( glsText ):
            gls[ m[0] ] = m[1:]

def _replace_function( m, tex, prevI, newTex ):
    a, b = m.span( )
    newTex += tex[prevI:a]
    return b, newTex

def replace_glossaries( tex, gls ):
    prevI, newTex, firstTime = 0, '', []
    glsPat = re.compile( r'\\gls\{\s*(?P<id>.+?)\}' )
    newTex, b = '', 0
    for m in glsPat.finditer( tex ):
        a, b = m.span( )
        prevI, newTex = _replace_function( m, tex, prevI, newTex )
        # Now find replacement.
        mm = '%s' % m.group(1)
        try:
            if mm not in firstTime:
                firstTime.append( mm )
                replaceWith = gls[ mm ][1] + ' (**%s**)' % gls[mm][0]
            else:
                replaceWith = gls[ mm ][0] 
            newTex += replaceWith
        except Exception as e:
            newTex += mm
    return newTex + tex[b:]

def unit_to_tex( unit ):
    replaceDict = { 
            'meter' : 'm', 'nanometer' : 'nm', 'kilometer' : 'km'
            , 'sec' : 'sec', 'second' : 's', 'hour' : 'h', 'minute' : 'min'
            , 'mole' : 'M', 'kilogram' : 'kg'
            }

    prefix = dict( per = '^-1^', square = '^2^', cubic = '^3^' )
    suffix = dict( squared = '^2^', cubed = '^3^' )
    unit = [ x.strip() for x in unit.split( '\\' ) if x.strip( ) ]

    newUnit = [ ]
    for i, u in enumerate( unit ):
        if u in replaceDict:
            # Now search for prefixes and suffixes.
            newU = replaceDict[u]
            for ii in range(i+1, len(unit) ):
                if unit[ii] in suffix:
                    newU += suffix[ unit[ii] ]
                    unit[ii] = ''
                else:
                    break
            for ii in range( i-1, -1, -1):
                if unit[ii] in prefix:
                    if unit[ii] == 'per':
                        if re.search( r'\^\d+\^', newU ):
                            newU = re.sub( r'\^(\d+)\^', r'^-\1^', newU )
                        else:
                            newU += '^-1^'
                    else:
                        newU += prefix[ unit[ii] ]
                    unit[ii] = ''
                else:
                    break
            newUnit.append( newU )
    return '{0}'.format( ''.join( newUnit ))


def si_replacement( val, unit ):
    if 'e' in val.lower( ):
        val = val.split( 'e' )
        val = r'%s×10^%s^' % (val[0],val[1])
    unit = unit_to_tex( unit )
    return '%s %s' % (val, unit)

def replace_si( text ):
    # yet to be implemented.
    pat = re.compile( r'\\SI\{(.+?)\}\{(.+?)}', re.DOTALL )
    siWithVals = pat.finditer( text )
    prevI, newText = 0, ''
    for m in siWithVals:
        a, b = m.span( )
        val, units = m.group(1), m.group(2)
        newText += text[prevI:a]
        replaceWith = si_replacement( val, units )
        #  print( '[INFO] %s %s %s' % (val,units,replaceWith.strip()))
        newText += replaceWith 
        prevI = b
    return newText + text[b:]

def read_glossaries( gls_files ):
    glossaries = { }
    for f in gls_files:
        read_glossary( f, glossaries )
    return glossaries

def replace( infile ):
    dirname = os.path.dirname( os.path.realpath(infile.name) )
    tex = infile.read( )
    glsFiles = find_glossaries( tex, dirname )
    gls = read_glossaries( glsFiles )
    text = replace_glossaries( tex, gls )
    text = replace_si( text )
    print( text, file = sys.stdout )

def main( args ):
    replace( args.INPUT )

if __name__ == '__main__':
    import argparse
    # Argument parser.
    description = '''Replace glossaries/acronyms (\gls) in TeX/Pandoc docs.'''
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('INPUT', nargs = '?'
            , type = argparse.FileType('r'), default = sys.stdin
            , help = 'Input TeX/pandoc file or stdin'
        )
    class Args: pass 
    args = Args()
    parser.parse_args(namespace=args)
    main( args )



__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

import sys
import dilawar.pandoc.utils as pu

def main():
    args = sys.argv[1:]
    if not args:
        args = [ '-h' ]
    pu.execute_pandoc(args)

if __name__ == '__main__':
    main()




#!/usr/bin/env python3

"""
Pandoc filter to process raw latex tikz environments into images.
Assumes that pdflatex is in the path, and that the standalone
package is available.  Also assumes that ImageMagick's convert
is in the path. Images are put in the tikz-images directory.

CREDIT: Did I write this? Or downloaded it from somewhere? Can't tell anymore.
But it has gone through substantial changes now. Please let me know if you feel
like it is derivative of your work.
"""

import os
import re
import shutil
import sys
import functools
import subprocess
import hashlib
import requests
import mimetypes
import glob

from pandocfilters import toJSONFilters, Para, Image, get_filename4code, get_extension
from tempfile import mkdtemp

# setup env.
import sys
sdir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(sdir)
import theorem 
import code_blocks

incomment = False

def log( *msg ):
    print( *msg, file = sys.stderr )

def get_filename( text ):
    m = hashlib.sha256( text.encode() ).hexdigest( )
    return '%s' % m

def download_image_from_url( url ):
    basename = get_filename4code( '_downloaded_from_url', url, '' )
    log( 'basename', basename )
    if os.path.isdir( basename ):
        # Return first file from this directory
        return  glob.glob( '%s/*' % basename )[0]
    try:
        r = requests.get( url, stream = True, timeout = 4)
    except Exception:
        return url

    ext = mimetypes.guess_extension( r.headers['content-type'] ) or url.split('.')[-1]
    if 'jpe' in ext:
        ext = 'jpg'

    os.makedirs( basename )
    filename = os.path.join( basename, f'downloaded_img.{ext}')
    if not os.path.exists( filename ):
        with open( filename, 'wb' ) as f:
            f.write( r.content )
    return filename

def tikz2image(tikz_src, filetype, outfile):
    tmpdir = mkdtemp()
    olddir = os.getcwd()
    os.chdir(tmpdir)

    # Write tikz.tex file.
    with open('tikz.tex', 'w') as f:
        f.write( '\n'.join( 
            [ "\\RequirePackage{luatex85,shellesc}"
                , "\\documentclass{standalone}"
                , "\\usepackage{tikz}"
                , "\\usepackage[sfdefault]{firasans}"
                , "\\usepackage[small,euler-digits]{eulervm}"
                , "\\usepackage{pgfplots}"
                , "\\pgfplotslibrary[]{units,groupplots}"
                , "\\begin{document}" ] 
            ))
        f.write(tikz_src)
        f.write("\n\\end{document}\n")

    subprocess.run( ["latexmk", "-pdf", "-lualatex", '--shell-escape', '-silent', 'tikz.tex']
            , stdout=sys.stderr
            )
    os.chdir(olddir)
    if filetype == 'pdf':
        shutil.copyfile(tmpdir + '/tikz.pdf', outfile + '.pdf')
    else:
        subprocess.run(["convert", tmpdir + '/tikz.pdf', outfile + '.' + filetype])
    shutil.rmtree(tmpdir)

def tikz(key, value, format, _):
    if key == 'RawBlock':
        [fmt, code] = value
        if fmt == "latex" and re.match( r'\begin{tikzpicture}', code):
            outfile = get_filename4code("tikz", code)
            filetype = get_extension(format, "png", html="png", latex="pdf")
            src = outfile + '.' + filetype
            if not os.path.isfile(src):
                tikz2image(code, filetype, outfile)
                log('Created image ' + src )
            return Para([Image(['', [], []], [], [src, ""])])

def comment(k, v, fmt, meta):
    global incomment
    if k == 'RawBlock':
        fmt, s = v
        if fmt == "html":
            if re.search("<!-- BEGIN COMMENT -->", s):
                incomment = True
                return []
            elif re.search("<!-- END COMMENT -->", s):
                incomment = False
                return []
    if incomment:
        return []  # suppress anything in a comment

def image_with_url( k, v, fmt, meta ):
    if k == 'Image':
        urlOrPath = v[2][0]
        if 'http' in urlOrPath:
            url = urlOrPath
            path = download_image_from_url( url )
            log( "[INFO ] Replacing url %s with downloaded file %s" % (url, path) )
            v[2][0] = path
            return Image( *v )

if __name__ == "__main__":
    toJSONFilters( 
        [ image_with_url, comment, theorem.theorems, tikz, code_blocks.codeblocks ] 
        ) 



"""data_analysis.py: 

Helper function related to data analysis.

"""
    
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2016, Dilawar Singh"
__credits__          = ["NCBS Bangalore"]
__license__          = "GNU GPL"
__version__          = "1.0.0"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"
__status__           = "Development"


import sys
import os
import math
import numpy as np
from collections import defaultdict
import scipy.signal  as _sig

def change_sign( vec ):
    # use np.signbit instead of np.sign
    # this is not the best way to compute the zero-crossing. 
    #
    # For b = [ 1, 2, 0, -1, 0, 0, -1, 2], it returns 3 crossing though there
    # are only 2.
    idx = np.signbit( vec ) != np.signbit( np.roll(vec, 1))
    return np.where( idx )[0][1:]

def isclose( v1, v2, eabs = 1e-3 ):
    return abs( v1 - v2 ) < eabs

def is_periodic_helper( vec ):
    # List of timeperiods. It may contain 1 timeperiod of multiple timeperid.
    # Essentially we are looking of repeating substring.
    tps = vec.copy( )
    res = [ ]
    while sum(tps) > 1:
        x0 = tps[np.nonzero(tps)][0]
        ts = [ i for i, x in enumerate(tps) if isclose(x,x0,max(3,0.1*x0)) ]
        if len(ts) < 2:
            continue
        res.append( (x0, np.mean(np.diff(ts))) )
        tps[ts] = 0

    periods, periodicity = zip( *res )
    if min(periodicity) == max(periodicity):
        return np.sum( periods )
    return 0

def find_period( vec, polar = False, ax = None):
    # normalize the vector
    if polar:
        vec[:,0] = vec[:,0] % (2*np.pi) 
    ref = vec[-1]
    dist = np.zeros( len(vec) )
    phase = np.zeros( len(vec) )
    periods = [ ]
    prevI = 0
    for i, v in enumerate(vec[::-1]):
        dist[i] = sum( (v - ref) ** 2) ** 0.5
        phase[i] = math.atan(v[1]/v[0]) - math.atan(ref[1] / ref[0])

    # normalize distance.
    if dist.max( ) == dist.min():
        return False

    if phase.max() == phase.min():
        return False

    if ax is not None:
        ax.plot( dist ) #, color = 'red' )

    zeroCross = change_sign( phase )
    valsAtZero = dist[zeroCross]

    # remove zero-crossings where distance is not close to zero.
    goodPoints = [ x for x in zip(zeroCross, valsAtZero) if not isclose(x[1],0) ]
    if len( goodPoints) < 1:
        return False

    zeroCross, valsAtZero = zip(*goodPoints)
    if len(zeroCross) == 0:
        return 0

    timePeriod = np.diff( zeroCross )
    return is_periodic_helper( timePeriod )


def smooth( sig, N = 100 ):
    window = np.ones( N ) / float( N )
    return np.convolve(  sig, window, 'same' )

def digitize( sig, levels, thres = 4 ):
    for x in levels:
        sig[ (sig > (x-thres)) & (sig < (x+thres)) ] = x
    sig[ np.isnan( sig ) ] = 0
    return sig 

def find_transitions( vec, levels, thres = 4 ):
    sig = digitize( vec, levels, thres )
    result = defaultdict( list )
    for x in levels:
        xIds = np.where( vec ==  x )[0]
        result[ 'kramer_time' ].append( (x, len( xIds ) / 1.0 / len( vec )) )

    trans = np.diff( sig ) 

    result[ 'up_transitions' ] = np.where( trans > 0 )[0]
    result[ 'down_transitions' ] = np.where( trans < 0 )[0]

    return result, sig

def test( datafile ):
    import matplotlib.pyplot as plt
    import pandas as pd
    data = pd.read_csv( datafile, sep = ' ', comment = '#' )
    data.dropna( how = 'any' )
    camkii = data.filter( regex = r'x0y\d.+' )
    tvec = data[ 'time' ]
    lowCaMKII = np.sum( camkii, axis = 1 )
    sig = smooth( lowCaMKII, 500 )
    res, newY = find_transitions( sig, [0,8,16] )
    print( res )
    upT = tvec[ res[ 'up_transitions' ] ]
    downT = tvec[ res[ 'down_transitions' ] ]

    plt.subplot( 311 )
    plt.plot( tvec, lowCaMKII  )
    plt.plot( upT, sig.max() * np.ones( len( upT ) ) , '+' )
    plt.plot( downT, sig.min() * np.ones( len( downT ) ) , '+' )
    plt.subplot( 312 )
    #plt.plot( tvec, step_detection( yvec, 0, 8 ) )
    plt.plot( tvec, sig )
    plt.subplot( 313 )
    plt.plot( tvec, newY )
    plt.savefig( '%s_transitions.png' % datafile )

def main( ):
    datafile = sys.argv[1]
    test(datafile)

# Alias. Deprecated
compute_transitions = find_transitions

if __name__ == '__main__':
    main()



# This file is originall from https://github.com/jtambasco/gnuplotpy

import os
import time
import shutil as sh
import numpy as np
import subprocess
import re
import tempfile
import matplotlib as mpl
import matplotlib.pyplot as plt
from . import nx_utils
from . import plot_utils
import dilawar.matplotlib_basic_units as mpl_basic_units

init_pgfplots_ = False


def _read_line(filename, line_number):
    s = None
    with open(filename, 'r') as fs:
        for i, line in enumerate(fs.readlines()):
            if i == line_number:
                s = line
    return s


class _GnuplotDeletingFile:
    def __init__(self, filename):
        self.name = filename

    def __del__(self):
        os.remove(self.name)


class _GnuplotScriptTemp(_GnuplotDeletingFile):
    def __init__(self, gnuplot_cmds):
        _GnuplotDeletingFile.__init__(self, '.tmp_gnuplot.gpi')
        with open(self.name, 'w') as fs:
            fs.write(gnuplot_cmds)


class _GnuplotDataTemp(_GnuplotDeletingFile):
    def __init__(self, *args):
        _GnuplotDeletingFile.__init__(self, '.tmp_gnuplot_data.dat')
        data = np.array(args).T
        with open(self.name, 'wb') as fs:
            np.savetxt(fs, data, delimiter=',')


class _GnuplotDataZMatrixTemp(_GnuplotDeletingFile):
    def __init__(self, z_matrix):
        _GnuplotDeletingFile.__init__(self, '.tmp_gnuplot_data_z_matrix.dat')
        with open(self.name, 'wb') as fs:
            np.savetxt(fs, z_matrix, '%.3f', delimiter=',')


#def gnuplot( script, args_dict={}, data=[]):
def gnuplot(script, **kwargs):
    '''
    Call a Gnuplot script, passing it arguments and
    datasets.

    Args:
        script(str): The name of the Gnuplot script or the text
        args_dict(dict): A dictionary of parameters to pass
            to the script.  The `key` is the name of the variable
            that the `item` will be passed to the Gnuplot script
            with.
        data(list): A list of lists containing lists to be plotted.
            The lists can be accessed by plotting the variable
            `data` in the Gnuplot script.  The first list in the
            list of lists corresponds to the first column in data,
            and so on.
    Returns:
        str: The Gnuplot command used to call the script.
    '''

    if os.path.exists(script):
        with open(script) as f:
            script = f.read()

    for k in kwargs:
        v = kwargs[k]
        script = script.replace('@%s@' % k, v)

    # Find rest of the macros.
    for m in re.findall(r'@\S+?@', script):
        script = script.replace(m, kwargs.get(m.replace('@', ''), ''))

    # if first argument is a long string then save the string to current
    # directory before running the command.
    # First escape all char.
    script = script.replace(r'"', r"'").replace(';', '')
    script = ';'.join(filter(None, script.split('\n')))
    script += ';exit;'
    scriptName = '.gnuplot_script'
    with open(scriptName, 'w') as f:
        f.write(script)

    while not os.path.isfile(scriptName):
        time.sleep(0.0001)

    subprocess.Popen(['gnuplot', scriptName])
    return True


def nx_draw(graph, ax=None, **kwargs):
    """Draw to PNG using graphviz (default = neato).

    Args:
    ----
        graph: networkx graph.
        ax: matplotlib axis. If `None`, `plt.gca()` is used.
        kwargs: Optional options to pass to `nx.draw_networkx` function.

    Return:
    ------
        'pos' computed by graphviz. 
    """
    import networkx as nx
    # pydot is pure python and easy to install. PyGraphviz Agraph interface
    # requires graphviz development libraries.
    try:
        from networkx.drawing.nx_agraph import graphviz_layout
    except ImportError as e:
        from networkx.drawing.nx_pydot import graphviz_layout
    if ax is None:
        ax = plt.gca()

    if kwargs.get('pos', None) is None:
        pos = graphviz_layout(graph, prog=kwargs.get('program', 'neato'))
    else:
        pos = kwargs['pos']

    if 'pos' in kwargs:
        del kwargs['pos']

    nx.draw_networkx(graph, pos=pos, ax=ax, **kwargs)
    # Draw edge labels.
    if kwargs.get('edge_labels', None) is not None:
        el = kwargs['edge_labels']
        if isinstance(el, str):
            elDict = {}
            for s, t in graph.edges():
                elDict[(s, t)] = graph[s][t].get(el, '')
        elif isinstance(el, dict):
            elDict = el
        else:
            elDict = {}
        nx.draw_networkx_edge_labels(graph, pos, edge_labels=elDict)
    return pos


def nx_draw_subprocess(graph, program='neato', ax=None):
    """Draw to PNG using graphviz (default = neato).
    """
    import matplotlib.image as mpimg
    # pydot is pure python and easy to install. PyGraphviz Agraph interface
    # requires graphviz development libraries.
    from networkx.drawing.nx_pydot import write_dot
    fh, dotfile = tempfile.mkstemp()
    pngfile = '%s.png' % dotfile
    write_dot(graph, dotfile)
    if ax is not None:
        subprocess.check_output([program, "-Tpng", dotfile, "-o", pngfile],
                                shell=False)
        if os.path.exists(pngfile):
            im = mpimg.imread(pngfile)
            ax.imshow(im, interpolation='none')
        else:
            raise UserWarning('Failed to draw graph using %s' % program)

def _matrix_plot(img, xvec, yvec, ax=None, **kwargs):
    if ax is None:
        ax = plt.subplot(111)
    img = np.matrix(img)
    nc, nr = img.shape
    im = ax.imshow(img,
                   interpolation=kwargs.get('interpolation', 'none'),
                   aspect=kwargs.get('aspect', 'auto'),
                   cmap=kwargs.get('cmap', 'viridis'))

    # apply ticks and labels
    xticks = kwargs.get('xticks', [])
    yticks = kwargs.get('yticks', [])
    if not xticks:
        nticks = kwargs.get('num_xticks', kwargs.get('num_ticks', 5))
        xticks = [(i, xvec[int(i)])
                for i in np.linspace(0, len(xvec), nticks)[:-1]]

    xpos, xlabels = zip(*xticks)
    ax.set_xticks(xpos)
    ax.set_xticklabels([r'%s' % x for x in xlabels])

    if not yticks:
        nticks = kwargs.get('num_yticks', kwargs.get('num_ticks', 5))
        yticks = [(i, yvec[int(i)])
                for i in np.linspace(0, len(yvec), nticks)[:-1]
                ]
        yticks.append((len(yvec)-1, yvec[-1]))

    ypos, ylabels = zip(*yticks)
    ax.set_yticks(ypos)
    ax.set_yticklabels([r'%.2g'%y for y in ylabels])

    ax.set_xlabel(kwargs.get('xlabel', 'NA'))
    ax.set_ylabel(kwargs.get('ylabel', 'NA'))
    ax.set_title(kwargs.get('title', ''))

    if kwargs.get('colorbar', True):
        plt.colorbar(im, ax=ax, format=kwargs.get('colorbar_fmt', '%g'))

    return im


def matrixPlot(img, xvec, yvec, ax=None, **kwargs):
    _matrix_plot(img, xvec, yvec, ax, **kwargs)


def matrix_plot(img, xvec, yvec, ax=None, **kwargs):
    raise DeprecationWarning("Use matrixPlot instead.")


def init_pgfplots(**kwargs):
    global init_pgfplots_
    if init_pgfplots_:
        return
    #  mathStyle = kwargs.get('math_style', 'upright')
    mpl.use('pgf')
    mpl.rcParams['text.latex.preamble'] = [
        r'\usepackage{mathtools}',
        r'\usepackage{libertine}',
        r'\usepackage{newtxmath}',
        r'\usepackage{tikz}'
        #, r'\usepackage[sfdefault,scale=0.9]{FiraSans}'
        #, r'\usepackage[default,scale=0.9]{opensans}'
        ,
        r'\usepackage[small,euler-digits]{eulervm}'
    ]
    mpl.rcParams['pgf.preamble'] = ''.join(mpl.rcParams['text.latex.preamble'])
    mpl.rcParams['text.usetex'] = True
    mpl.rcParams['lines.linewidth'] = 1

    # The following settings allow you to select the fonts in math mode.
    mpl.rcParams['mathtext.fontset'] = 'stixsans'
    mpl.rcParams['mathtext.default'] = 'regular'
    # AXES
    mpl.rcParams['axes.labelsize'] = 'medium'
    mpl.rcParams['axes.formatter.use_mathtext'] = True
    mpl.rcParams['axes.formatter.min_exponent'] = 0
    mpl.rcParams['axes.formatter.useoffset'] = True
    mpl.rcParams['axes.formatter.offset_threshold'] = 4

    mpl.rcParams['axes.spines.left'] = True
    mpl.rcParams['axes.spines.bottom'] = True
    mpl.rcParams['axes.spines.top'] = False
    mpl.rcParams['axes.spines.right'] = False

    mpl.rcParams['axes.autolimit_mode'] = 'data'
    mpl.rcParams['axes.xmargin'] = 0.1
    mpl.rcParams['axes.ymargin'] = 0.1

    # TICKS
    # see http://matplotlib.org/api/axis_api.html#matplotlib.axis.Tick
    mpl.rcParams['xtick.labelsize'] = 'small'
    mpl.rcParams['xtick.direction'] = 'inout'
    mpl.rcParams['ytick.labelsize'] = 'small'
    mpl.rcParams['ytick.direction'] = 'inout'

    # Legend
    mpl.rcParams['legend.loc'] = 'best'
    mpl.rcParams['legend.frameon'] = False
    mpl.rcParams['legend.framealpha'] = 0
    mpl.rcParams['legend.fancybox'] = True
    mpl.rcParams['legend.fontsize'] = 'small'
    mpl.rcParams['legend.borderpad'] = 0
    mpl.rcParams['legend.columnspacing'] = 1.0
    mpl.rcParams['legend.handlelength'] = 1.5
    init_pgfplots_ = True


def initPGFBackend(**kwargs):
    """Initialise PGF backend with preloaded packages."""
    init_pgfplots(**kwargs)
    return True


def pgfplots(df, xname, yname, ax, **kwargs):
    """Plot normal x-y curve like with pdfplots like settings.
    Also put valus into a dataframe and return it so it can be saved into 
    a csv file.
    """
    init_pgfplots()
    ax.plot(
        df[xname],
        df[yname]
        #  , kwargs.get('plot_style', '-')
        ,
        **kwargs.get('Line2D', {}))

    defaultLegendOptions = dict(bbox_to_anchor=(1, 1),
                                loc='upper left',
                                fontsize='x-small')
    defaultLegendOptions.update(kwargs.get('legend_option', {}))
    if kwargs.get('legend', ''):
        ax.legend(**defaultLegendOptions)

    if kwargs.get('xlabel', ''):
        ax.set_xlabel(kwargs['xlabel'])
    if kwargs.get('ylabel', ''):
        ax.set_ylabel(kwargs['ylabel'])
    if kwargs.get('title', ''):
        ax.set_title(kwargs.get('title', ''), fontsize='small')

    # Add label.
    if kwargs.get('label', ''):
        ax.text(-0.2,
                1.25,
                r'\textbf{%s}' % kwargs['label'],
                fontsize='medium',
                transform=ax.transAxes)


def phase_plot(x, y, ax):
    import math
    print("[WARN ] This function has not been tested.")
    ax.plot(x, y)
    X, Y = np.meshgrid(x, y)
    U, V = np.zeros(X.shape), np.zeros(Y.shape)
    for i, xx in enumerate(x):
        for j, yy in enumerate(y):
            l = (xx**2 + yy**2)**0.5
            theta = math.atan2(yy, xx)
            U[i, j] = xx + l * math.cos(theta)
            V[i, j] = yy + l * math.sin(theta)
    ax.quiver(X, Y, U, V)


def simple_axis(ax):
    # Create a simple axis.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()


def addLabel(label, ax, **kwargs):
    # Add label to axis.
    x, y = kwargs.get('x', -0.2), kwargs.get('y', 1.15)
    ax.text(x, y, label, transform=ax.transAxes)


def mid(x: np.array):
    """Find midpoint of given array.
    Useful to find midpoint of bins returned by np.histogram function.
    """
    return (x[:-1] + x[1:]) / 2


def _test_nx_draw(ax):
    import io
    import networkx as nx
    f = io.StringIO('''digraph G {
    a -> b [weight=1];
    a -> c [weight=2];
    }''')
    g = nx.DiGraph(nx_utils.read_dot(f))
    assert g.number_of_edges() == 2
    assert g.number_of_nodes() == 3
    plot_utils.nx_draw(g, ax=ax, edge_labels='weight')


def test():
    ax1 = plt.subplot(121)
    ax2 = plt.subplot(122)
    phase_plot([-1, 0, 1, 2, 3], [3, 1, 0.1, 0.2, 0.4], ax1)
    _test_nx_draw(ax2)
    plt.savefig('%s.png' % __file__)


if __name__ == '__main__':
    test()



# Hopfield netowrk demo using
__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"

import sys
import os
import matplotlib.pyplot as plt
import numpy as np




#!/usr/bin/env python3
"""
Glossaries in markdown should be written as \gls{entry}.

There should be a master file specified somewhere in YAML file as `glossaries`.
The format of this file is same as LaTeX glossaries package.
"""

import os
import sys
import re
import panflute as P

gls_ = {}
first_ = []

def _log(*args):
    print(*args, file=sys.stderr)

def readGlossaries(gfile):
    global gsl_
    glpat = re.compile( r'\\newacronym\{\s*(?P<gls>.+?)\}' + \
            r'\s*\{(?P<short>.+?)\}\s*\{(?P<long>.+?)\}', re.DOTALL )

    with open(gfile, 'r') as h:
        glsText = h.read()
        for m in glpat.findall(glsText):
            gls_[m[0]] = m[1:]

def replaceGls(text, gls):
    global first_
    m = re.match(r'\\gls{(\S+)}', text)
    if m:
        key = m.group(1)
        if key in first_:
            if key in gls:
                return gls[key][0]
            else:
                return text
        else:
            first_.append(key)
            if key in gls:
                return gls[key][1] + f' (**{gls[key][0]}**)'
            else:
                return text
    else:
        return text

def prepare_gls(doc):
    glsFile = doc.get_metadata('glossaries')
    # Thanks to snippet here
    # https://github.com/sergiocorreia/panflute/issues/74#issue-308922983
    if doc.format in ['latex']:
        return
    if glsFile is None or not glsFile.strip():
        return
    if os.path.exists(glsFile.strip()):
        readGlossaries(glsFile.strip())

def finalize_gls(doc):
    glsFile = doc.get_metadata('glossaries')
    if glsFile is None:
        return
    if doc.format in ['latex']:
        lines =[r'\usepackage[acronym]{glossaries}'
                , r'\loadglsentries{%s}'%glsFile]
        tex = [P.MetaInlines(P.RawInline(l, format='latex')) for l in lines]
        tex = P.MetaList(*tex)
        if 'header-includes' not in doc.metadata:
            doc.metadata['header-includes'] = tex
        else:
            doc.metadata['header-includes'].content.extend(tex)

def action_gls(elem, doc):
    global gls_
    if isinstance(elem, P.RawInline):
        if not gls_:
            return
        if '\\gls' in elem.text:
            new = P.convert_text(replaceGls(elem.text, gls_))[0]
            if isinstance(new, P.Para):
                return list(new.content)
            return P.Str(new.text)

def main(doc=None):
    P.run_filter(action_gls, prepare=prepare_gls, finalize=finalize_gls, doc=doc)

if __name__ == '__main__':
    main()



import os
import datetime
from setuptools import setup

# __version__ will be evaluated.
with open('./dilawar/version.py') as f:
    exec(f.read())

version_ = __version__
stamp = datetime.datetime.now().strftime('%Y%m%d')
if os.environ.get('TRAVIS'):
    version_ += '-dist%s' % stamp

print( "[INFO ] Packing %s" % version_ )

with open("README.md") as f:
    readme = f.read()

install_requires = ['pandoc-imagine', 'pathlib']
with open('requirements.txt') as f:
    install_requires = [x.strip() for x in f.read().split('\n') if x.strip()]

setup(
    name = "dilawar",
    version = version_,
    description = "My personal utilities. See the README.md file.",
    long_description = readme,
    long_description_content_type='text/markdown',
    packages = [ "dilawar", 'dilawar.pandoc' ],
    package_dir = { "dilawar" : 'dilawar' },
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.tex', '*.css', 'template/*'],
        },
    install_requires = install_requires, 
    author = "Dilawar Singh",
    author_email = "dilawars@ncbs.res.in",
    url = "http://github.com/dilawar/dilawar",
    license='GPLv3'
)



__author__           = "Dilawar Singh"
__copyright__        = "Copyright 2019-, Dilawar Singh"
__maintainer__       = "Dilawar Singh"
__email__            = "dilawars@ncbs.res.in"


def read_dot(dotfile):
    import networkx as nx
    try:
        return nx.drawing.nx_adot.read_dot(dotfile)
    except Exception as e:
        return nx.drawing.nx_pydot.read_dot(dotfile)

def test():
    import tempfile
    import io
    dot = io.StringIO('''
digraph D {
    a -> b;
    b -> c;
    } ''')
    g = read_dot(dot)
    assert g.number_of_nodes() == 3
    assert g.number_of_edges() == 2

def main():
    test()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
from pandocfilters import walk, toJSONFilter
from caps import caps

"""
Pandoc filter that causes emphasized text to be displayed
in ALL CAPS.
"""


def deemph(key, val, fmt, meta):
    if key == 'Emph':
        return walk(val, caps, fmt, meta)

if __name__ == "__main__":
    toJSONFilter(deemph)
